from django.contrib import messages
from django.urls import reverse
from django.views.generic import FormView

from arkindex.images.forms import ImageServerMergeForm, ImageServerSplitForm
from arkindex.images.models import ImageServer


class ImageServerAdminView(FormView):

    def get_context_data(self, **kwargs):
        data = super().get_context_data(**kwargs)
        data.update(
            object=self.server,
            opts=self.model_admin.opts,
            **self.model_admin.admin_site.each_context(self.request),
        )
        return data

    def get_success_url(self):
        return reverse("admin:images_imageserver_changelist")

    @property
    def server(self):
        return ImageServer.objects.get(id=self.kwargs["id"])


class ImageServerMergeView(ImageServerAdminView):
    template_name = "admin/images/imageserver/merge.html"
    form_class = ImageServerMergeForm

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["servers"] = ImageServer.objects.exclude(id=self.kwargs["id"])
        return kwargs

    def form_valid(self, form):
        source = self.server
        destination = form.cleaned_data["destination"]

        try:
            source.merge_into(destination)
        except ValueError as e:
            self.model_admin.message_user(
                self.request,
                f"Merging failed: {str(e)}",
                level=messages.ERROR,
            )
            return super().form_valid(form)

        source.delete()

        self.model_admin.message_user(
            self.request,
            f"Successfully merged {source} into {destination}",
            level=messages.SUCCESS,
        )
        return super().form_valid(form)


class ImageServerSplitView(ImageServerAdminView):
    template_name = "admin/images/imageserver/split.html"
    form_class = ImageServerSplitForm

    def form_valid(self, form):
        folder = form.cleaned_data["folder"]

        try:
            new_server = self.server.split(folder)
        except ValueError as e:
            self.model_admin.message_user(
                self.request,
                f"Splitting failed: {str(e)}",
                level=messages.ERROR,
            )
            return super().form_valid(form)

        self.model_admin.message_user(
            self.request,
            f'Successfully split "{folder}" from {self.server} into {new_server}',
            level=messages.SUCCESS,
        )
        return super().form_valid(form)
