from datetime import datetime, timezone

from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import Agent, Farm
from arkindex.process.models import (
    FeatureUsage,
    ProcessMode,
    Worker,
    WorkerRun,
    WorkerVersion,
)
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestWorkerRunsRetrieve(FixtureAPITestCase):
    """
    Test worker runs retrieve endpoint
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.local_process = cls.user.processes.get(mode=ProcessMode.Local)
        cls.farm = Farm.objects.first()
        cls.process_1 = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Workers,
            farm=cls.farm,
        )
        cls.version_1 = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_1 = cls.version_1.worker

        cls.worker_custom = Worker.objects.get(slug="custom")
        cls.version_custom = cls.worker_custom.versions.get()
        cls.run_1 = cls.process_1.worker_runs.create(version=cls.version_1)
        cls.run_custom = cls.local_process.worker_runs.get(version=cls.version_custom)
        cls.process_2 = cls.corpus.processes.create(creator=cls.user, mode=ProcessMode.Workers)

        cls.agent = Agent.objects.create(
            farm=cls.farm,
            fingerprint="a" * 64,
            hostname="claude",
            cpu_cores=42,
            cpu_frequency=1e15,
            ram_total=99e9,
            last_ping=datetime.now(timezone.utc),
        )
        # Add custom attributes to make the agent usable as an authenticated user
        cls.agent.is_agent = True
        cls.agent.is_anonymous = False

    def test_retrieve_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retrieve_no_worker_execution_right(self):
        """
        A user can retrieve any worker run if they have an admin access to the process project
        """
        self.worker_1.memberships.update(level=Role.Guest.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": None,
            "model_version": None,
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image_iid": self.version_1.docker_image_iid,
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision_url": None,
                "version": 1,
                "tag": None,
                "branch": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer",
                    "description": "",
                    "repository_url": None,
                    "archived": False,
                    "cost_1k_elements": "0.250",
                    "cost_cpu_hour": "1.312",
                    "cost_gpu_hour": "0.000",
                }
            },
            "use_gpu": False,
            "summary": "Worker Recognizer @ version 1",
        })

    def test_retrieve_invalid_id(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:worker-run-details", kwargs={"pk": "12341234-1234-1234-1234-123412341234"})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_retrieve(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": None,
            "model_version": None,
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image_iid": self.version_1.docker_image_iid,
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision_url": None,
                "version": 1,
                "tag": None,
                "branch": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer",
                    "description": "",
                    "repository_url": None,
                    "archived": False,
                    "cost_1k_elements": "0.250",
                    "cost_cpu_hour": "1.312",
                    "cost_gpu_hour": "0.000",
                }
            },
            "use_gpu": False,
            "summary": "Worker Recognizer @ version 1",
        })

    def test_retrieve_custom(self):
        self.client.force_login(self.user)
        # Update process attributes to ensure they do not cause extra requests
        page = self.corpus.elements.get(name="Volume 1, page 1r")
        self.local_process.element = page
        self.local_process.element_type = page.type
        self.local_process.folder_type = page.type
        self.local_process.prefix = "a"
        self.local_process.save()
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:worker-run-details", kwargs={"pk": str(self.run_custom.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.run_custom.id),
            "configuration": None,
            "model_version": None,
            "parents": [],
            "process": {
                "id": str(self.local_process.id),
                "activity_state": "disabled",
                "corpus": None,
                "chunks": 1,
                "mode": "local",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": {
                    "id": str(page.id),
                    "type": "page",
                    "corpus": {
                        "id": str(self.corpus.id),
                        "name": "Unit Tests",
                        "public": True
                    },
                    "name": "Volume 1, page 1r",
                    "rotation_angle": 0,
                    "mirrored": False,
                    "thumbnail_url": None,
                    "zone": {
                        "id": str(page.id),
                        "image": {
                            "id": str(page.image.id),
                            "height": 1000,
                            "path": "img1",
                            "s3_url": None,
                            "server": {
                                "display_name": "Test Server",
                                "max_height": None,
                                "max_width": None,
                                "url": "http://server"
                            },
                            "status": "unchecked",
                            "url": "http://server/img1",
                            "width": 1000
                        },
                        "polygon": [[int(x), int(y)] for x, y in page.polygon],
                        "url": "http://server/img1/0,0,1000,1000/full/0/default.jpg",
                    },
                },
                "element_type": "page",
                "folder_type": "page",
                "prefix": "a",
            },
            "worker_version": {
                "id": str(self.version_custom.id),
                "configuration": {"custom": "value"},
                "docker_image_iid": None,
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision_url": None,
                "version": 1,
                "tag": None,
                "branch": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "created",
                "worker": {
                    "id": str(self.worker_custom.id),
                    "name": "Custom worker",
                    "slug": "custom",
                    "type": "custom",
                    "description": "",
                    "repository_url": None,
                    "archived": False,
                    "cost_1k_elements": "0.000",
                    "cost_cpu_hour": "0.000",
                    "cost_gpu_hour": "0.000",
                }
            },
            "summary": "Worker Custom worker @ version 1",
            "use_gpu": False,
        })

    def test_retrieve_local(self):
        """
        A user can retrieve a run on their own local process
        """
        run = self.local_process.worker_runs.create(version=self.version_1)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(run.id),
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image_iid": self.version_1.docker_image_iid,
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision_url": None,
                "version": 1,
                "tag": None,
                "branch": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer",
                    "description": "",
                    "repository_url": None,
                    "archived": False,
                    "cost_1k_elements": "0.250",
                    "cost_cpu_hour": "1.312",
                    "cost_gpu_hour": "0.000",
                }
            },
            "parents": [],
            "model_version": None,
            "configuration": None,
            "process": {
                "id": str(self.local_process.id),
                "activity_state": "disabled",
                "corpus": None,
                "chunks": 1,
                "mode": "local",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "use_gpu": False,
            "summary": "Worker Recognizer @ version 1",
        })

    def test_retrieve_local_only_current_user(self):
        """
        A user cannot retrieve a run on another user's local process
        """
        run = WorkerRun.objects.filter(process__creator=self.superuser, process__mode=ProcessMode.Local).first()
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_retrieve_agent(self):
        """
        A Ponos agent can retrieve a WorkerRun on a process where it has some assigned tasks
        """
        self.process_1.tasks.create(run=0, depth=0, slug="something", agent=self.agent, ttl=0)

        # Agent auth is not implemented in CE
        self.client.force_authenticate(user=self.agent)
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "id": str(self.run_1.id),
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image_iid": self.version_1.docker_image_iid,
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision_url": None,
                "version": 1,
                "tag": None,
                "branch": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer",
                    "description": "",
                    "repository_url": None,
                    "archived": False,
                    "cost_1k_elements": "0.250",
                    "cost_cpu_hour": "1.312",
                    "cost_gpu_hour": "0.000",
                }
            },
            "parents": [],
            "model_version": None,
            "configuration": None,
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "mode": "workers",
                "chunks": 1,
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "use_gpu": False,
            "summary": "Worker Recognizer @ version 1",
        })

    def test_retrieve_agent_unassigned(self):
        """
        A Ponos agent cannot retrieve a WorkerRun on a process where it does not have any assigned tasks
        """
        self.process_1.tasks.create(run=0, depth=0, slug="something", agent=None, ttl=0)

        # Agent auth is not implemented in CE
        self.client.force_authenticate(user=self.agent)
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
